{% block doc -%}
<!DOCTYPE html>
<html{% block html_attribs %}{% endblock html_attribs %}>
  {%- block html %}
  <head>
    {%- block head %}
    <title>{% block title %}{{ title|default }}{% endblock title %}</title>

      {%- block metas %}
      <meta name="viewport" content="width=device-width, initial-scale=1.0">
      {%- endblock metas %}

      {%- block styles %}
      {%- endblock styles %}
    {%- endblock head %}
  </head>
  <body{% block body_attribs %}{% endblock body_attribs %}>
    {% block body -%}
      {% block navbar %}
      {%- endblock navbar %}
      {% block content -%}
      {%- endblock content %}

      {% block scripts %}
      {%- endblock scripts %}
    {%- endblock body %}
  </body>
  {%- endblock html %}
</html>
{% endblock doc -%}
